@@ -1,10 +1,13 @@
 from functools import cache
+from json import dumps
+from json import loads
+from os import W_OK
+from os import access
 from pathlib import Path
 from re import match
 from re import split
 from re import sub
 from sqlite3 import DatabaseError
-from typing import Optional
 
 from falocalrepo_database import Column
 from falocalrepo_database import Database as _Database
@@ -71,7 +74,7 @@ def query_to_sql(query: str, default_field: str, likes: list[str] = None, aliase
 class Database(_Database):
     def __init__(self, database_path: Path):
         _Database.check_connection(database_path)
-        super().__init__(database_path, read_only=True)
+        super().__init__(database_path, read_only=not access(database_path, W_OK))
         if not self.is_formatted:
             raise DatabaseError("Database not formatted")
 
@@ -125,17 +128,34 @@ class Database(_Database):
         return self.journals[journal_id]
 
     @cache
-    def _load_prev_next_cached(self, table: str, item_id: int, *, _cache=None) -> tuple[int, int]:
-        table: Table = self[table]
-        item: dict | None = table[item_id]
-        query: Selector = Sb("AUTHOR").__eq__(item["AUTHOR"])
-        query = Sb() & [query, Sb("FOLDER").__eq__(item["FOLDER"])] if table == submissions_table else query
-        return table.select(
-            query,
-            columns=[Column("LAG(ID, 1, 0) over (order by ID)", int), Column("LEAD(ID, 1, 0) over (order by ID)", int)],
-            order=[f"ABS(ID - {item_id})"],
-            limit=1
-        ).cursor.fetchone() if item else (0, 0)
+    def _load_prev_next_cached(self, table: str, item_id: int | str, *, _cache=None) -> tuple[int, int]:
+        db_table: Table = self[table]
+
+        if table in (submissions_table, journals_table):
+            if not (item := db_table[item_id]):
+                return 0, 0
+
+            query: Selector = Sb("AUTHOR").__eq__(item["AUTHOR"])
+            query = Sb() & [query, Sb("FOLDER").__eq__(item["FOLDER"])] if table.upper() == submissions_table else query
+            return db_table.select(
+                query,
+                columns=[Column("LAG(ID, 1, 0) over (order by ID)", int),
+                         Column("LEAD(ID, 1, 0) over (order by ID)", int)],
+                order=[f"ABS(ID - {item_id})"],
+                limit=1
+            ).cursor.fetchone()
+        elif table == users_table:
+            if not (item := db_table[item_id]):
+                return 0, 0
+
+            query1: Selector = Sb("USERNAME").__gt__(item["USERNAME"])
+            query2: Selector = Sb("USERNAME").__lt__(item["USERNAME"])
+            return (
+                next(db_table.select(query1, columns=["USERNAME"], order=["USERNAME ASC"], limit=1).cursor, [0])[0],
+                next(db_table.select(query2, columns=["USERNAME"], order=["USERNAME DESC"], limit=1).cursor, [0])[0]
+            )
+        else:
+            raise KeyError(f"Unknown table {table!r}")
 
     @cache
     def _load_search_cached(self, table: str, query: str, sort: str, order: str, *, _cache=None):
@@ -150,7 +170,7 @@ class Database(_Database):
                             SubmissionsColumns.DATE.value, SubmissionsColumns.TITLE.value]
             db_table = self.submissions if table == submissions_table else self.journals
         else:
-            cols_results = [UsersColumns.USERNAME.value, UsersColumns.FOLDERS.value]
+            cols_results = [UsersColumns.USERNAME.value, UsersColumns.FOLDERS.value, UsersColumns.ACTIVE.value]
             db_table = self.users
 
         cols_table: list[str] = [c.name for c in db_table.columns]
@@ -159,8 +179,14 @@ class Database(_Database):
 
         sql, values = query_to_sql(query,
                                    default_field,
-                                   [*map(str.lower, {*cols_table, "any"} - {"ID", "AUTHOR", "USERNAME"})],
+                                   [*map(str.lower,
+                                         {*cols_table, "any", "keywords", "message", "filename"} -
+                                         {"ID", "AUTHOR", "USERNAME", "FILESAVED", "USERUPDATE", "ACTIVE"})],
                                    {"author": "replace(author, '_', '')",
+                                    "lower": "replace(author, '_', '')",
+                                    "keywords": "tags",
+                                    "message": "description",
+                                    "filename": "fileurl",
                                     "any": f"({'||'.join(cols_table)})"},
                                    score=sort == "relevance")
 
@@ -170,7 +196,7 @@ class Database(_Database):
                 dict(zip([c.name for c in cols_results] + ["RELEVANCE"], s)) for s in
                 db_table.select_sql(f"RELEVANCE > 0", values,
                                     columns=[*cols_results, Column(f"({sql if sql else 1}) as RELEVANCE", int)],
-                                    order=[f"{sort} {order}", f"{default_sort[table]} {default_order[table]}"]).cursor]
+                                    order=[f"{sort} {order}", f"{default_sort[table]} {default_order[table]}"]).tuples]
         else:
             results = list(db_table.select_sql(sql, values, columns=cols_results, order=[f"{sort} {order}"]))
         return (
@@ -192,6 +218,16 @@ class Database(_Database):
         return len(self.users), len(self.submissions), len(
             self.journals), self.version
 
+    def save_settings(self, name: str, settings: dict):
+        if self.read_only:
+            return
+
+        self.settings[f"SERVER.{name}"] = dumps(settings)
+        self.commit()
+
+    def load_settings(self, name: str) -> dict:
+        return loads(self.settings[f"SERVER.{name}"] or "{}")
+
     def load_user(self, user: str) -> dict | None:
         return self._load_user_cached(user, _cache=self.m_time)
 
@@ -207,7 +243,7 @@ class Database(_Database):
     def load_journal(self, journal_id: int) -> dict | None:
         return self._load_journal_cached(journal_id, _cache=self.m_time)
 
-    def load_prev_next(self, table: str, item_id: int) -> tuple[int, int]:
+    def load_prev_next(self, table: str, item_id: int | str) -> tuple[int, int]:
         return self._load_prev_next_cached(table, item_id, _cache=self.m_time)
 
     def load_search(self, table: str, query: str, sort: str, order: str):
@@ -234,7 +270,7 @@ class Database(_Database):
     def load_journal_uncached(self, journal_id: int) -> dict | None:
         return self._load_journal_cached.__wrapped__(self, journal_id)
 
-    def load_prev_next_uncached(self, table: str, item_id: int) -> tuple[int, int]:
+    def load_prev_next_uncached(self, table: str, item_id: int | str) -> tuple[int, int]:
         return self._load_prev_next_cached.__wrapped__(self, table, item_id)
 
     def load_search_uncached(self, table: str, query: str, sort: str, order: str):
