@@ -6,10 +6,12 @@ from logging import Logger
 from logging import getLogger
 from os import PathLike
 from pathlib import Path
+from re import IGNORECASE
 from re import Pattern
 from re import compile as re_compile
 from secrets import compare_digest
 from sqlite3 import DatabaseError
+from sqlite3 import OperationalError
 from typing import Any
 from typing import Callable
 from typing import Coroutine
@@ -17,7 +19,12 @@ from zipfile import ZipFile
 
 from PIL import Image
 from PIL import UnidentifiedImageError
+from bs4 import BeautifulSoup
+from bs4.element import Tag
 from chardet import detect as detect_encoding
+from falocalrepo_database.tables import JournalsColumns
+from falocalrepo_database.tables import SubmissionsColumns
+from falocalrepo_database.tables import UsersColumns
 from fastapi import FastAPI
 from fastapi import Request
 from fastapi import Response
@@ -33,6 +40,7 @@ from fastapi.security import HTTPBasic
 from fastapi.security import HTTPBasicCredentials
 from fastapi.staticfiles import StaticFiles
 from fastapi.templating import Jinja2Templates
+from htmlmin import minify
 from pydantic import BaseModel
 from pydantic import BaseSettings
 from uvicorn import run
@@ -48,6 +56,25 @@ from .database import submissions_table
 from .database import users_table
 
 
+class SearchSettings(BaseSettings):
+    view: dict[str, str] = {users_table: "grid", submissions_table: "grid", journals_table: "grid"}
+    limit: dict[str, str] = {users_table: "48", submissions_table: "48", journals_table: "48"}
+    sort: dict[str, str] = default_sort
+    order: dict[str, str] = default_order
+
+    def reset(self):
+        self.view = {users_table: "grid", submissions_table: "grid", journals_table: "grid"}
+        self.limit = {users_table: "48", submissions_table: "48", journals_table: "48"}
+        self.sort = default_sort
+        self.order = default_order
+
+    def load(self, obj: dict):
+        self.view = obj.get("view", self.view)
+        self.limit = obj.get("limit", self.limit)
+        self.sort = obj.get("sort", self.sort)
+        self.order = obj.get("order", self.order)
+
+
 class Settings(BaseSettings):
     database: Database = None
     static_folder: Path = None
@@ -71,13 +98,15 @@ LOGGING_CONFIG["formatters"]["access"]["fmt"] = \
     '%(levelprefix)s %(asctime)s %(client_addr)s - %(request_line)s %(status_code)s %(msecs).0fms'
 
 fa_base_url: str = "https://www.furaffinity.net"
+fa_link: Pattern = re_compile(r"(https?://)?(www.)?furaffinity.net", flags=IGNORECASE)
 root: Path = Path(__file__).resolve().parent
 app: FastAPI = FastAPI(title="FurAffinity Local Repo", openapi_url=None)
 templates: Jinja2Templates = Jinja2Templates(str(root / "templates"))
 settings: Settings = Settings(static_folder=root / "static")
+search_settings: SearchSettings = SearchSettings()
 security: HTTPBasic = HTTPBasic()
 
-tags_expressions: list[tuple[Pattern, str]] = [
+bbcode_expressions: list[tuple[Pattern, str]] = [
     (re_compile(r"\[(/?)([bius]|sup|sub|h\d)]"), r"<\1\2>",),
     (re_compile(r"\[color=([^]]+)]"), r'<span style="color: \1">,'),
     (re_compile(r"\[/color]"), "</span>",),
@@ -90,41 +119,63 @@ tags_expressions: list[tuple[Pattern, str]] = [
     (re_compile(r"\[/url]"), "</a>"),
     (re_compile(r"\[yt]((?:.(?!\[yt]))+)\[/yt]"), r'</a href="\1">\1</a>'),
     (re_compile(r"(:icon([^:]+):|:([^:]+)icon:)"), r'<a href="/user/\2\3">@\2\3</a>'),
+    (re_compile(r"((?:\n(?!\n)|[^\n])+)\n\n+"), r"<p>\1</p>"),
+    (re_compile(r"\n"), "<br/>")
 ]
 
-button: Callable[[str, str], str] = lambda h, t: f'<a href="{h}"><button>{t}</button></a>'
-
 app.mount("/static", StaticFiles(directory=settings.static_folder), "static")
 
 
-def tags_to_html(text: str) -> str:
-    return reduce(lambda t, es: es[0].sub(es[1], t), tags_expressions, text)
+def bbcode_to_html(text: str) -> str:
+    return reduce(lambda t, es: es[0].sub(es[1], t), bbcode_expressions, text.strip() + "\n\n")
+
+
+def clean_html(html: str) -> str:
+    html_parsed: BeautifulSoup = BeautifulSoup(html, "lxml")
+    for icon in html_parsed.select("a.iconusername > img"):
+        parent: Tag = icon.parent
+        parent.clear()
+        parent.insert(0, f"@{icon.attrs['title']}")
+    for code in html_parsed.select("code"):
+        code.name = "div"
+    for link in html_parsed.select("a[href*='furaffinity.net']"):
+        link.attrs["href"] = "/" + fa_link.sub("", link.attrs["href"]).strip("/")
+    return str(html_parsed)
 
 
 def serialise_entry(entry: dict) -> dict:
     return {k: sorted(v) if isinstance(v, set) else str(v) if isinstance(v, datetime) else v for k, v in entry.items()}
 
 
+def error_response(request: Request, code: int, message: str = None, buttons: list[tuple[str, str]] = None) -> Response:
+    return templates.TemplateResponse(
+        "error.html",
+        {"app": app.title,
+         "title": f"Error {code}",
+         "code": code,
+         "message": message,
+         "buttons": buttons,
+         "request": request},
+        code
+    )
+
+
 async def auth_middleware(request: Request, call_next: Callable[[Request], Coroutine[Any, Any, Response]]) -> Response:
+    if request.url.path.startswith("/static"):
+        return await call_next(request)
+
     try:
         creds: HTTPBasicCredentials = await security(request)
         if compare_digest(creds.username, settings.username) and compare_digest(creds.password, settings.password):
             return await call_next(request)
     except HTTPException as err:
         if err.status_code != status.HTTP_401_UNAUTHORIZED:
-            return serve_error(request, err.detail, err.status_code)
-    return Response("Incorrect username or password", status.HTTP_401_UNAUTHORIZED, {"WWW-Authenticate": "Basic"})
+            return error_response(request, err.status_code, err.detail)
 
-
-def serve_error(request: Request, message: str, code: int) -> Response:
-    return templates.TemplateResponse(
-        "error.html",
-        {"title": f"{app.title} · Error {code}",
-         "code": code,
-         "message": message,
-         "request": request},
-        code
-    )
+    return Response(
+        error_response(request, status.HTTP_401_UNAUTHORIZED, "Incorrect username or password",
+                       [("Retry", str(request.url))]).body,
+        status.HTTP_401_UNAUTHORIZED, {"WWW-Authenticate": "Basic"})
 
 
 @app.on_event("startup")
@@ -166,14 +217,14 @@ async def error_unknown(request: Request, err: HTTPException):
     logger.error(repr(err))
     if request.method == "POST":
         return JSONResponse({"errors": [{err.__class__.__name__: err.detail}]}, err.status_code)
-    return serve_error(request, err.detail or err.__class__.__name__, err.status_code)
+    return error_response(request, err.status_code, err.detail or None)
 
 
 @app.exception_handler(404)
 async def error_not_found(request: Request, err: HTTPException):
     if request.method == "POST":
         return JSONResponse({"errors": [{err.__class__.__name__: err.detail}]}, err.status_code)
-    return serve_error(request, err.detail or "Not found", err.status_code)
+    return error_response(request, err.status_code, err.detail or "Not found")
 
 
 @app.exception_handler(422)
@@ -182,32 +233,41 @@ async def error_not_found(request: Request, err: RequestValidationError):
     logger.error(f"{err.__class__.__name__} {err.errors()}")
     if request.method == "POST":
         return JSONResponse({"errors": err.errors()}, 422)
-    return serve_error(request, err.errors()[0].get("msg", None) or err.__class__.__name__, 422)
+    return error_response(request, 422, err.errors()[0].get("msg", None) or err.__class__.__name__)
 
 
 @app.exception_handler(DatabaseError)
-async def error_database(request: Request, err: DatabaseError):
+@app.exception_handler(OperationalError)
+async def error_database(request: Request, err: DatabaseError | OperationalError):
     logger.error(repr(err))
     if request.method == "POST":
         return JSONResponse({"errors": [{err.__class__.__name__: err.args}]}, 500)
-    return serve_error(request, err.__class__.__name__ + "<br/>" + "<br/>".join(err.args or []), 500)
+    return error_response(request, 500, "<br/>".join([err.__class__.__name__, *map(str, err.args)]))
 
 
-@app.get("/view/{id_}", response_class=HTMLResponse)
-@app.get("/full/{id_}", response_class=HTMLResponse)
+@app.exception_handler(FileNotFoundError)
+async def error_database_not_found(request: Request, err: FileNotFoundError):
+    logger.error(repr(err))
+    if request.method == "POST":
+        return JSONResponse({"errors": [{err.__class__.__name__: err.args}]}, status.HTTP_503_SERVICE_UNAVAILABLE)
+    return error_response(request, 500, "Database not found")
+
+
+@app.get("/view/{id_}/", response_class=HTMLResponse)
+@app.get("/full/{id_}/", response_class=HTMLResponse)
 async def redirect_submission(id_: int):
     return RedirectResponse(app.url_path_for(serve_submission.__name__, id_=str(id_)),
                             status.HTTP_301_MOVED_PERMANENTLY)
 
 
-@app.get("/gallery/{username}", response_class=HTMLResponse)
+@app.get("/gallery/{username}/", response_class=HTMLResponse)
 @app.get("/search/gallery/{username}/", response_class=HTMLResponse)
 async def serve_user_gallery(request: Request, username: str):
     return await serve_search(request, "submissions", f"Gallery {username}",
                               {"query": f'@author "{clean_username(username)}" @folder "gallery"'})
 
 
-@app.get("/scraps/{username}", response_class=HTMLResponse)
+@app.get("/scraps/{username}/", response_class=HTMLResponse)
 @app.get("/search/scraps/{username}/", response_class=HTMLResponse)
 async def serve_user_scraps(request: Request, username: str):
     return await serve_search(request, "submissions", f"Scraps {username}",
@@ -228,56 +288,99 @@ async def serve_user_journals(request: Request, username: str):
                               {"query": f'@author "{clean_username(username)}"'})
 
 
-@app.get("/favorites/{username}", response_class=HTMLResponse)
+@app.get("/favorites/{username}/", response_class=HTMLResponse)
 @app.get("/search/favorites/{username}/", response_class=HTMLResponse)
 async def serve_user_favorites(request: Request, username: str):
     return await serve_search(request, "submissions", f"Favorites {username}",
                               {"query": f'@favorite "|{clean_username(username)}|"'})
 
 
-@app.get("/mentions/{username}", response_class=HTMLResponse)
+@app.get("/mentions/{username}/", response_class=HTMLResponse)
 @app.get("/search/mentions/{username}/", response_class=HTMLResponse)
 async def serve_user_mentions(request: Request, username: str):
     return await serve_search(request, "submissions", f"Mentions {username}",
                               {"query": f'@mentions "|{clean_username(username)}|"'})
 
 
+@app.get("/search/")
+async def server_search_default(request: Request):
+    return RedirectResponse(app.url_path_for(serve_search.__name__, table=submissions_table.lower()) +
+                            "?" + request.url.query,
+                            status.HTTP_301_MOVED_PERMANENTLY)
+
+
 @app.get("/", response_class=HTMLResponse)
 async def serve_home(request: Request):
     usr_n, sub_n, jrn_n, version = settings.database.load_info()
-    return templates.TemplateResponse(
-        "info.html",
-        {"title": app.title,
-         "submissions_total": sub_n,
-         "journals_total": jrn_n,
-         "users_total": usr_n,
-         "version_db": version,
-         "version": __version__,
-         "request": request}
-    )
+    return HTMLResponse(minify(templates.get_template("info.html").render({
+        "app": app.title,
+        "title": "",
+        "submissions_total": sub_n,
+        "journals_total": jrn_n,
+        "users_total": usr_n,
+        "version_db": version,
+        "version": __version__,
+        "request": request}),
+        remove_comments=True))
 
 
-@app.get("/user/{username}", response_class=HTMLResponse)
+@app.get("/settings/")
+async def serve_settings(request: Request):
+    return HTMLResponse(minify(templates.get_template("settings.html").render({
+        "app": app.title,
+        "title": "Search Settings",
+        "tables": [users_table, submissions_table, journals_table],
+        "columns": {users_table: [c.name for c in UsersColumns],
+                    submissions_table: [c.name for c in SubmissionsColumns] + ["relevance"],
+                    journals_table: [c.name for c in JournalsColumns]},
+        "settings": search_settings,
+        "default_sort": default_sort,
+        "default_order": default_order,
+        "request": request}),
+        remove_comments=True))
+
+
+@app.get("/settings/set/")
+async def save_settings(request: Request):
+    for param, value in request.query_params.items():
+        table, setting = param.split(".")
+        search_settings.__setattr__(setting, search_settings.__getattribute__(setting) | {table.upper(): value})
+
+    if settings.database.read_only:
+        return Response("Database is read only, settings will be reset on restart",
+                        status_code=status.HTTP_403_FORBIDDEN)
+
+    settings.database.save_settings("SEARCH", search_settings.dict())
+
+    return Response("Settings saved", status_code=200)
+
+
+@app.get("/user/{username}/", response_class=HTMLResponse)
 async def serve_user(request: Request, username: str):
     if username != (username_clean := clean_username(username)):
         return RedirectResponse(app.url_path_for(serve_user.__name__, username=username_clean))
 
     user_entry: dict | None = settings.database.load_user(username)
     user_stats: dict[str, int] = settings.database.load_user_stats(username)
+    p, n = settings.database.load_prev_next(users_table, username) if user_entry else (0, 0)
 
-    return templates.TemplateResponse(
-        "user.html",
-        {"title": f"{app.title} · {username}",
-         "user": username,
-         "folders": user_entry["FOLDERS"] if user_entry else [],
-         "gallery_length": user_stats["gallery"],
-         "scraps_length": user_stats["scraps"],
-         "favorites_length": user_stats["favorites"],
-         "mentions_length": user_stats["mentions"],
-         "journals_length": user_stats["journals"],
-         "userpage": user_entry["USERPAGE"],
-         "request": request}
-    )
+    return HTMLResponse(minify(templates.get_template("user.html").render({
+        "app": app.title,
+        "title": username,
+        "user": username,
+        "folders": user_entry["FOLDERS"] if user_entry else [],
+        "active": user_entry["ACTIVE"] if user_entry else True,
+        "gallery_length": user_stats["gallery"],
+        "scraps_length": user_stats["scraps"],
+        "favorites_length": user_stats["favorites"],
+        "mentions_length": user_stats["mentions"],
+        "journals_length": user_stats["journals"],
+        "userpage": clean_html(user_entry["USERPAGE"]) if user_entry else "",
+        "in_database": user_entry is not None,
+        "prev": p,
+        "next": n,
+        "request": request}),
+        remove_comments=True))
 
 
 @app.get("/search/{table}/", response_class=HTMLResponse)
@@ -287,16 +390,14 @@ async def serve_search(request: Request, table: str, title: str = None, args: di
 
     args = {k.lower(): v for k, v in (args or {}).items()}
     args_req = {k.lower(): v for k, v in request.query_params.items()}
-    query: str = " & ".join([f"({q})" for args_ in (args_req, args) if (q := args_.get("query", None))])
+    query: str = " & ".join([f"({q})" for args_ in (args_req, args) if (q := args_.get("query", args_.get("q", None)))])
     args |= args_req
 
     page: int = p if (p := int(args.get("page", 1))) > 0 else 1
-    limit: int = l if (l := int(args.get("limit", 48))) > 0 else 48
-    sort: str = args.get("sort", default_sort[table]).lower()
-    order: str = args.get("order", default_order[table]).lower()
-    view: str = args.get("view", "").lower()
-    view = "grid" if view not in ("list", "grid") and table == submissions_table else view
-    view = "list" if table != submissions_table else view
+    limit: int = l if (l := int(args.get("limit", search_settings.limit[table]))) > 0 else 48
+    sort: str = args.get("sort", search_settings.sort[table]).lower()
+    order: str = args.get("order", search_settings.order[table]).lower()
+    view: str = v if (v := args.get("view", search_settings.view[table]).lower()) in ("list", "grid") else "grid"
 
     results: list[dict]
     columns_table: list[str]
@@ -311,52 +412,50 @@ async def serve_search(request: Request, table: str, title: str = None, args: di
         order
     )
 
-    return templates.TemplateResponse(
-        "search.html",
-        {"title": f"{app.title} · " + (title or f"{request.url.path.split('/')[1].title()} {table.title()}"),
-         "action": request.url.path,
-         "table": table.lower(),
-         "query": query,
-         "sort": sort,
-         "order": order,
-         "view": view,
-         "allow_view": table == submissions_table,
-         "thumbnails": table == submissions_table,
-         "columns_table": columns_table,
-         "columns_results": columns_results,
-         "columns_list": columns_list,
-         "column_id": column_id,
-         "limit": limit,
-         "page": page,
-         "offset": (offset := (page - 1) * limit),
-         "results": results[offset:offset + limit],
-         "results_total": len(results),
-         "request": request}
-    )
+    return HTMLResponse(minify(templates.get_template("search.html").render({
+        "app": app.title,
+        "title": title or f"Search {table.title()}",
+        "action": request.url.path,
+        "table": table.lower(),
+        "query": args_req.get("query", args_req.get("q", "")),
+        "sort": sort,
+        "order": order,
+        "view": view,
+        "thumbnails": table == submissions_table,
+        "columns_table": columns_table,
+        "columns_results": columns_results,
+        "columns_list": columns_list,
+        "column_id": column_id,
+        "limit": limit,
+        "page": page,
+        "offset": (offset := (page - 1) * limit),
+        "results": results[offset:offset + limit],
+        "results_total": len(results),
+        "request": request}),
+        remove_comments=True))
 
 
 @app.get("/submission/{id_}/", response_class=HTMLResponse)
 async def serve_submission(request: Request, id_: int):
     if (sub := settings.database.load_submission(id_)) is None:
-        raise HTTPException(
-            404,
-            f"Submission not found.<br>{button(f'{fa_base_url}/view/{id_}', 'Open on Fur Affinity')}", )
+        return error_response(request, 404, "Submission not found",
+                              [("Open on Fur Affinity", f"{fa_base_url}/view/{id_}")])
 
     f: Path | None = None
     if sub["FILEEXT"] == "txt" and sub["FILESAVED"] & 0b10:
         f, _ = settings.database.load_submission_files(id_)
     p, n = settings.database.load_prev_next(submissions_table, id_)
-    return templates.TemplateResponse(
-        "submission.html",
-        {"title": f"{app.title} · {sub['TITLE']} by {sub['AUTHOR']}",
-         "submission": sub,
-         "file_text": tags_to_html(f.read_text(encoding=detect_encoding(f.read_bytes())["encoding"])) if f else None,
-         "filename": f"submission{('.' + sub['FILEEXT']) * bool(sub['FILEEXT'])}",
-         "filename_id": f"{sub['ID']:010d}{('.' + sub['FILEEXT']) * bool(sub['FILEEXT'])}",
-         "prev": p,
-         "next": n,
-         "request": request}
-    )
+    return HTMLResponse(minify(templates.get_template("submission.html").render({
+        "app": app.title,
+        "title": f"{sub['TITLE']} by {sub['AUTHOR']}",
+        "submission": sub | {"DESCRIPTION": clean_html(sub["DESCRIPTION"])},
+        "file_text": bbcode_to_html(f.read_text(encoding=detect_encoding(f.read_bytes())["encoding"])) if f else None,
+        "filename": f"submission{('.' + sub['FILEEXT']) * bool(sub['FILEEXT'])}",
+        "filename_id": f"{sub['ID']:010d}{('.' + sub['FILEEXT']) * bool(sub['FILEEXT'])}",
+        "prev": p,
+        "next": n,
+        "request": request}),
+        remove_comments=True))
 
 
 @app.get("/submission/{id_}/file/")
@@ -418,19 +517,18 @@ async def serve_submission_zip(id_: int, _filename=None):
 @app.get("/journal/{id_}/", response_class=HTMLResponse)
 async def serve_journal(request: Request, id_: int):
     if (jrnl := settings.database.load_journal(id_)) is None:
-        raise HTTPException(
-            404,
-            f"Journal not found.<br>{button(f'{fa_base_url}/journal/{id_}', 'Open on Fur Affinity')}")
+        return error_response(request, 404, "Journal not found",
+                              [("Open on Fur Affinity", f"{fa_base_url}/journal/{id_}")])
 
     p, n = settings.database.load_prev_next(journals_table, id_)
-    return templates.TemplateResponse(
-        "journal.html",
-        {"title": f"{app.title} · {jrnl['TITLE']} by {jrnl['AUTHOR']}",
-         "journal": jrnl,
-         "prev": p,
-         "next": n,
-         "request": request}
-    )
+    return HTMLResponse(minify(templates.get_template("journal.html").render({
+        "app": app.title,
+        "title": f"{jrnl['TITLE']} by {jrnl['AUTHOR']}",
+        "journal": jrnl | {"CONTENT": clean_html(jrnl["CONTENT"])},
+        "prev": p,
+        "next": n,
+        "request": request}),
+        remove_comments=True))
 
 
 @app.get("/journal/{id_}/zip/")
@@ -498,8 +596,8 @@ async def serve_journal_json(id_: int):
         return j
 
 
-@app.get("/json/user/{username}", response_class=JSONResponse)
-@app.post("/json/user/{username}", response_class=JSONResponse)
+@app.get("/json/user/{username}/", response_class=JSONResponse)
+@app.post("/json/user/{username}/", response_class=JSONResponse)
 async def serve_user_json(username: str):
     username = clean_username(username)
     user_entry: dict | None = settings.database.load_user_uncached(username)
@@ -540,5 +638,6 @@ def server(database_path: str | PathLike, host: str = "0.0.0.0", port: int = Non
         run_args |= {"port": port or 443, "ssl_certfile": settings.ssl_cert, "ssl_keyfile": settings.ssl_key}
     run_args |= {"port": run_args.get("port", port) or 80}
     with Database(Path(database_path).resolve()) as settings.database:
+        search_settings.load(settings.database.load_settings("SEARCH"))
         # noinspection PyTypeChecker
         run(app, host=host, **run_args)
